@@ -2,6 +2,8 @@
 # vim:tabstop=2
 # vim:shiftwidth=2
 # vim:enc=utf-8
+# vim:foldmethod=marker
+# vim:foldenable
 package Net::Vypress::Chat;
 
 use 5.008;
@@ -16,13 +18,19 @@ require Exporter;
 
 our @ISA = qw(Exporter);
 
-our $VERSION = '0.62';
+our $VERSION = '0.70';
 
 # Prints debug messages
 sub debug { # {{{
 	my ($self, $text, $buffer) = @_;
 	print "*** $text\n" if $self->{debug};
-	print "($buffer)\n" if $self->{'debug'} == 2;
+	if ($buffer && $self->{debug} == 2) {
+		my $header = substr $buffer, 0, 1;
+		my $random = substr $buffer, 1, 9;
+		my $left = substr $buffer, 10;
+		$left =~ s/\0/|/gs;
+		print "($header $random $left)\n";
+	}
 } # }}}
 
 # Generates Vypress Chat header used to mark its packets.
@@ -163,6 +171,33 @@ sub here_ack { # {{{
 		.num2active($self->{'users'}{$self->{'nick'}}{'active'}), $str);
 } # }}}
 
+# Sends string thru unicast
+sub usend { # {{{
+	my ($self, $str, $to) = @_;
+	if (defined $self->{users}{$to}{ip}) {
+		my $iaddr = inet_aton($self->{users}{$to}{ip});
+		my $paddr = sockaddr_in($self->{port}, $iaddr);
+		$self->debug("F: usend, To: $to, Ip: $self->{users}{$to}{ip}", $str);
+		$self->{usend}->send($str, 0, $paddr);
+	}
+	elsif ($self->{uc_fail} == 1) {
+		$self->debug("F: usend, To: $to, Warn: IP unknown, A: Sending bcast.");
+		$self->{send}->send($str);
+	}
+	else {
+		$self->debug("F: usend, To: $to, Err: IP unknown");
+	}
+} # }}}
+
+# Sends string thru usend to people on some chan
+sub usend_chan { # {{{
+	my ($self, $str, $chan) = @_;
+	$self->debug("F: usend_chan, Chan: $chan");
+	for (@{$self->{channels}{$chan}{users}}) {
+		$self->usend($str, $_);
+	}	
+} # }}}
+
 
 
 # {{{ Documentation start
@@ -315,7 +350,19 @@ be warned in console. Also $vyc->{badip} variable will be set to 1.
 
 =back
 
-E.g.: $vyc->{'debug'}=2;
+=item uc_fail
+- toggles sending thru broadcast socket when unicast socket fails (ip cannot be
+found). Default: 1.
+
+=over
+
+=back
+
+=item *
+0 - drop packet.
+
+=item *
+1 - send packet thru broadcast socket instead of unicast.
 
 =back
 
@@ -334,6 +381,7 @@ sub new { # {{{
 	$self->{nick} = "default";
 	$self->{port} = $args{port} || 8167;
 	$self->{debug} = $args{debug} || 0;
+	$self->{uc_fail} = $args{uc_fail} || 1;
 	$self->{send_info} = (defined $args{send_info}) ? $args{send_info} : 1;
 	$self->{sign_topic} = (defined $args{sign_topic}) ? $args{sign_topic} : 1;
 	$self->{host} = $args{host} || hostname();
@@ -374,7 +422,8 @@ sub init_users { # {{{
 			'active'	=>	$tmpactive,
 			'gender'	=>	$tmpgender,
 			'autoanswer'	=>	$tmpaa,
-			'chats' => @tmpchats
+			'chats' => @tmpchats,
+			'ip' => $self->{localip},
 		}
 	};
 	$self->debug("init_users(), Status: "
@@ -531,7 +580,7 @@ sub remote_exec { # {{{
 	my ($self, $to, $command, $password) = @_;
 	my $str = header()."8".$self->{nick}."\0".$to."\0".$command."\0".$password
 		."\0";
-	$self->{'send'}->send($str);
+	$self->usend($str, $to);
 	$self->debug("Sent remote execution request to $to:\n"
 		."Password: $password\n"
 		."Command line: $command\n", $str);
@@ -548,7 +597,7 @@ E.g.: $vyc->remote_exec_ack('OtherGuy', 'Some text');
 sub remote_exec_ack { # {{{
 	my ($self, $to, $text) = @_;
 	my $str = header()."9".$to."\0".$self->{nick}."\0".$text."\0";
-	$self->{send}->send($str);
+	$self->usend($str, $to);
 	$self->debug("Sent remote execution acknowledgement to $to:\n"
 		."Execution text: $text", $str);
 } # }}}
@@ -564,7 +613,7 @@ E.g.: $vyc->sound_req("#Main", 'clap.wav');
 sub sound_req { # {{{
 	my ($self, $chan, $file) = @_;
 	my $str = header()."I".$self->{nick}."\0".$file."\0".$chan."\0";
-	$self->{send}->send($str);
+	$self->usend_chan($str, $chan);
 	$self->debug("Sent sound request for file $file to $chan", $str);
 } # }}}
 
@@ -579,7 +628,7 @@ E.g.: $vyc->me("#Main", "jumps around.");
 sub me { # {{{
 	my ($self, $chan, $text) = @_;
 	my $str = header()."A".$chan."\0".$self->{nick}."\0".$text."\0";
-	$self->{send}->send($str);
+	$self->usend_chan($str, $chan);
 	$self->debug("Did /me action in $chan: $text", $str);
 } # }}}
 
@@ -594,7 +643,7 @@ E.g.: $vyc->chat("#Main", "Hello!");
 sub chat { # {{{
 	my ($self, $chan, $text) = @_;
 	my $str = header()."2".$chan."\0".$self->{'nick'}."\0".$text."\0";
-	$self->{'send'}->send($str);
+	$self->usend_chan($str, $chan);
 	$self->debug("Sent chat string to $chan: $text", $str);
 } # }}}
 
@@ -612,7 +661,7 @@ sub join { # Join to channel {{{
 		my $str = header()."4".$self->{'nick'}."\0".$chan."\0"
 			.$self->{'users'}{$self->{'nick'}}{'status'}
 			.$self->{'users'}{$self->{'nick'}}{'gender'};
-		$self->{'send'}->send($str);
+		$self->usend_chan($str, $chan);
 		$self->add_to_channel($self->{nick}, $chan);
 		$self->debug("F: join(), Chan: $chan", $str);
 	}
@@ -634,7 +683,7 @@ sub part { # {{{
 	if ($self->on_chan($self->{nick}, $chan)) {
 		my $str = header()."5".$self->{'nick'}."\0".$chan."\0"
 			.$self->{'users'}{$self->{'nick'}}{'gender'};
-		$self->{'send'}->send($str);
+		$self->usend_chan($str, $chan);
 
 		$self->delete_from_channel($self->{nick}, $chan);
 		$self->debug("F: part(), Chan: $chan", $str);
@@ -658,7 +707,7 @@ sub topic { # {{{
 	$signature = ' ('.$self->{'nick'}.')' if $topic && $self->{sign_topic};
 	my $str = header()."B".$chan."\0".$topic.$signature."\0";
 	$self->{'channels'}{$chan}{'topic'} = $topic;
-	$self->{'send'}->send($str);
+	$self->usend_chan($str, $chan);
 	$self->debug("F: topic(), Chan: $chan, Topic: \"$topic\"", $str);
 } # }}}
 
@@ -674,9 +723,8 @@ E.g.: $vyc->msg("John", "Hello there...");
 sub msg { # {{{
 	my ($self, $to, $msg) = @_;
 	my $str = header()."6".$self->{'nick'}."\0".$to."\0".$msg."\0";
-	$self->{'send'}->send($str);
+	$self->usend($str, $to);
 	$self->debug("Sent msg for $to: \"$msg\"", $str);
-	return 0;
 } # }}}
 
 =head2 mass($message)
@@ -692,7 +740,7 @@ sub mass { # {{{
 	for (keys %{$self->{'users'}}) {
 		unless ($_ eq $self->{'nick'}) {
 			my $str = header()."E".$self->{'nick'}."\0".$_."\0".$msg."\0";
-			$self->{'send'}->send($str);
+			$self->usend($str, $_);
 			$self->debug("F: mass(), To: $_, Text: \"$msg\"", $str);
 		}
 		else {
@@ -701,6 +749,26 @@ sub mass { # {{{
 	}
 } # }}}
 
+=head2 mass_to(@to, $message)
+
+Sends message to  people in array. The message is marked as multi-user message.
+
+E.g.: $vyc->mass(('John', 'Paul'), "Hi everyone, I'm back.");
+
+=cut
+
+sub mass_to { # {{{
+	my $self = shift;
+	my $msg = pop;
+	my @to = @_;
+	for (@to) {
+		my $str = header()."E".$self->{'nick'}."\0".$_."\0".$msg."\0";
+		$self->usend($str, $_);
+		$self->debug("F: mass_to(), To: $_, Text: \"$msg\"", $str);
+	}
+} # }}}
+
+
 =head2 status($status, $autoanswer)
 
 Changes your status into one of four states mentioned in new() and
@@ -762,7 +830,7 @@ E.g.: $vyc->beep('OtherGuy');
 sub beep { # {{{
 	my ($self, $to) = @_;
 	my $str = header()."H0".$to."\0".$self->{send}."\0";
-	$self->{send}->send($str);
+	$self->usend($str, $to);
 	$self->debug("F: beep(), To: $to", $str);
 } # }}}
 
@@ -792,7 +860,7 @@ E.g.: $vyc->info("John");
 sub info { # {{{
 	my ($self, $to) = @_;
 	my $str = header()."F".$to."\0".$self->{'nick'}."\0";
-	$self->{'send'}->send($str);
+	$self->usend($str, $to);
 	$self->debug("F: info(), To: $to", $str);
 } # }}}
 
@@ -871,7 +939,7 @@ sub info_ack { # {{{
 	my $str = header() ."G". $to ."\0". $self->{'nick'} ."\0". $host
 		 ."\0". $user ."\0". $ip ."\0". $chans ."#\0"
 	   . $aa ."\0";
-	$self->{'send'}->send($str);
+	$self->usend($str, $to);
 	$self->debug("F: info_ack(), To: $to, Nick: $self->{'nick'}, Host: $host "
 		. "User: $user, IP: $ip, Chans: $chans, AA: $aa", $str);
 } # }}}
@@ -889,7 +957,7 @@ sub pjoin { # {{{
 	unless ($self->on_priv($to)) {
 		my $str = header() ."J0". $self->{nick} ."\0". $to ."\0"
 			. $self->{users}{$self->{nick}}{gender};
-		$self->{send}->send($str);
+		$self->usend($str, $to);
 		$self->add_to_private($to);
 		$self->debug("F: pjoin(), To: $to", $str);
 	}
@@ -911,7 +979,7 @@ sub ppart { # {{{
 	if ($self->on_priv($to)) {
 		my $str = header() ."J1". $self->{nick} ."\0". $to ."\0"
 			. $self->{users}{$self->{nick}}{gender};
-		$self->{send}->send($str);
+		$self->usend($str, $to);
 		$self->delete_from_private($to);
 		$self->debug("F: ppart(), To: $to", $str);
 	}
@@ -934,7 +1002,7 @@ sub pchat { # {{{
 	if ($self->on_priv($to)) {
 		my $str = header() ."J2". $self->{nick} ."\0". $to ."\0"
 			. $text ."\0";
-		$self->{send}->send($str);
+		$self->usend($str, $to);
 		$self->debug("F: pchat(), To: $to", $str);
 	}
 	else {
@@ -956,7 +1024,7 @@ sub pme { # {{{
 	if ($self->on_priv($to)) {
 		my $str = header() ."J3". $self->{nick} ."\0". $to ."\0"
 			. $text ."\0";
-		$self->{send}->send($str);
+		$self->usend($str, $to);
 		$self->debug("F: pme(), To: $to", $str);
 	}
 	else {
@@ -988,6 +1056,16 @@ sub startup { # {{{
 		LocalAddr => $self->{'localip'},
 		Broadcast => 1 ) || croak ("Failed! ($!)");
 	$self->debug("Success.");
+	# Outgoing unicast port.
+	$self->debug("Trying to open unicast socket from $self->{localip} to port "
+		."$self->{port}...");
+	$self->{'usend'} = IO::Socket::INET->new(
+		PeerPort => $self->{'port'},
+		Proto	=> 'udp',
+		LocalAddr => $self->{'localip'}
+		) || croak ("Failed! ($!)");
+	$self->debug("Success.");
+
 	# Incoming port.
 	$self->debug("Trying to estabilsh socket on $self->{localip}:"
 		."$self->{port}...");
@@ -1681,9 +1759,15 @@ __END__
 
 =back
 
+=head1 TRICKS
+
+=head2 Getting userlist for channel.
+
+Userlist for channel is stored in $vyc->{channels}{$chan}{users}. It's an array.
+
 =head1 SEE ALSO
 
-L<IO::Socket> L<IO::Socket::Inet> L<IO::Select> L<recv>
+L<IO::Socket> L<IO::Socket::Inet> L<IO::Select>
 
 Official web page of Vypress Chat: L<http://vypress.com/products/chat/>
 
